==> mutual_credit-5.0.5.tar.gz: mutual_credit/modules/clearingcentral/src/ClearingCentral.php <==
<?php

namespace Drupal\mcapi_cc;

use Drupal\mcapi\Entity\TransactionInterface;
use Drupal\mcapi\Entity\Wallet;
use Drupal\mcapi\Entity\Storage\WalletStorage;
use Drupal\mcapi\Plugin\Validation\Constraint\CanPayIn;
use Drupal\mcapi\Plugin\Validation\Constraint\CanPayOut;
use Drupal\mcapi\Entity\CurrencyInterface;
use Drupal\mcapi\Entity\Currency;
use Drupal\mcapi\Entity\Transaction;
use Drupal\user\Entity\User;
use Drupal\Core\Session\AccountSwitcher;
use Drupal\Core\Logger\LoggerChannel;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\KeyValueStore\KeyValueFactory;
use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Messenger\MessengerInterface;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\RequestException;

/**
 * Manage communications with clearing central web service
 */
class ClearingCentral implements IntertradingInterface {

  protected $nameLookup;
  protected $httpClient;
  protected $logger;
  protected $country;
  protected $accountSwitcher;
  protected $messenger;
  protected $walletEntityQuery;

  private $exchangeId;
  private $password;
  private $currencyId;

  /**
   * @var Drupal\user\Entity\User
   */
  private $localWalletOwner;


  const CLEARING_CENTRAL_URL = 'http://clearingcentral.net';
  const CLEARING_CENTRAL_IP = '69.61.35.151';

  const SUCCESS = 1;
  const FAILURE = 0;
  const ERR_INVALID_USER = 2; //user (buyer) does not exist (unknown account number)
  const ERR_EXCHANGE = 3;
  const ERR_DENIED = 4; //"no funds", over limit, account locked, exchange over deficit limit etc.
  const ERR_SYNTAX= 5; // Faulty data
  const ERR_DUPLICATE= 6; //same TXID submitted
  const ERR_URL = 7;
  const ERR_CONVERSION = 8;
  const ERR_SERVER = 9;
  const ERR_AUTH = 10;
  const ERR_UNKNOWN_IP = 11;
  const ERR_MISSING_TXID = 12; //update/delete only
  const ERR_UNKNOWN_TXID = 13; //update/delete only
  const ERR_REMOTE = 14; //Unable to connect to remote server
  const ERR_PEER_FAILED = 17; //Peer failed verification


  /**
   * @param ConfigFactory $config_factory
   * @param ClientInterface $http_client
   * @param KeyValueFactory $key_value
   * @param EntitytypeManager $entity_type_manager
   * @param LoggerChannel $logger_channel
   * @param AccountSwitcher $account_switcher
   * @param MessengerInterface $messenger
   */
  function __construct(ConfigFactory $config_factory, ClientInterface $http_client, KeyValueFactory $key_value, EntityTypeManager $entity_type_manager, LoggerChannel $logger_channel, AccountSwitcher $account_switcher, MessengerInterface $messenger) {
    $this->siteConfig = $config_factory->get('system.site');
    $this->httpClient  = $http_client;
    $this->nameLookup  = $key_value->get('exchangeNames');
    $this->walletEntityQuery = $entity_type_manager->getStorage('mcapi_wallet')->getQuery();
    $this->logger = $logger_channel;
    $this->country = $config_factory->get('system.date')->get('country.default');
    $this->accountSwitcher = $account_switcher;
    $this->messenger = $messenger;

    $settings = $config_factory->get('mcapi_cc.settings');
    $this->exchangeId = $settings->get('login');
    $this->pass = $settings->get('pass');
    $this->currencyId = $settings->get('curr_id');
  }

  /**
   * {@inheritdoc}
   */
  public function send(TransactionInterface $transaction, $remote_exchange_id, $remote_user_id) : float {
    $remote_transaction = $this->prepareToSend(
      $transaction,
      $transaction->outgoing,
      $remote_exchange_id,
      $remote_user_id
    );
    $this->logger->debug('<pre>@array</pre>', ['@array' =>  print_r($remote_transaction, 1)]);
    try {
      $url = self::CLEARING_CENTRAL_URL.'/txinput.php?'.http_build_query($remote_transaction);
      $this->logger->debug('@url', ['@url' => $url]);
      $result = $this->httpClient->post($url);
    }
    catch (RequestException $e) {
      $this->messenger->addError($e->getMessage());
      return 0;
    }

    if ($result->getStatusCode() == 200) {
      $contents = $result->getBody()->getContents();
      parse_str($contents, $params);
      $response_code = (int)$params['response'];
      switch ($response_code) {
        case SELF::SUCCESS:
          // These are needed in hook_mcapi_transaction_insert.
          $transaction->remote->exchange_id = $remote_exchange_id;
          $transaction->remote->user_id = $remote_user_id;
          return $params['amount'];
        default:
          $this->messenger->addError($this->responseLookup($response_code));
          break;
      }
    }
    else {
      $message = t('Clearing central failed to respond.');
      $this->logger->error('@string', ['@string' => $message]);
      $this->messenger->addError($message);
      $form_state->setRebuild(TRUE);
    }
    return 0;
  }


  /**
   * {@inheritdoc}
   */
  public function receive(array &$params) : void {
    $props = [
      'uuid' => $params['txid'],
      'type' => 'remote',
      'description' => html_entity_decode($params['description'])
    ];

    if ($params['outgoing']) {
      //payment is going from the originating exchange to this exchange
      $local_exchange_id = $params['seller_nid'];
      $other_exchange_id = $params['buyer_nid'];
      $other_exchange_name = $params['buyer_xname'];
      $other_user_name = $params['buyer_name'];
      $wid = $this->walletIdFromfragment($params['seller_id']);
    }
    else {
      //payment is going from this exchange to the originating exchange
      $local_exchange_id = $params['buyer_nid'];
      $other_exchange_id = $params['seller_nid'];
      $other_exchange_name = $params['seller_xname'];
      $other_user_name = $params['seller_name'];
      $wid = $this->walletIdFromfragment($params['buyer_id']);
    }
    if (!$wid) {
      $this->logger->error(
        t(
          'Unable to find one wallet for fragment @frag',
          ['@frag' => $params['buyer_id']]
        )
      );
      $params['response'] = SELF::ERR_INVALID_USER;
      return;
    }
    else {
      // Replace the given seller_id with the actual wallet name.
      $params['seller_id'] = Wallet::load($wid)->name;
    }
    //store this for when we need to view the transaction
    if ($other_exchange_name) {
      $this->nameLookup->set($other_exchange_id, $other_exchange_name);
    }
    list($parts[1], $parts[3]) = explode('.', number_format($params['amount'], 2));

    $props['worth'] = [
      [
        'curr_id' => $this->currencyId,
        'value' => Currency::load($this->currencyId)->unformat($parts)
      ]
    ];
    $props['creator'] = 0;

    // wallet owner was created above in walletIdFromfragment()
    $this->accountSwitcher->switchTo($this->localWalletOwner);

    if ($params['outgoing']) {
      $props['payer'] = intertrading_wallet_id();
      $props['payee'] = $wid;
      $holder = Wallet::load($props['payee'])->getHolder();
      $params['seller_name'] = $holder->label();
      $params['seller_email'] = $holder->getEmail();
      $params['seller_xname'] = $this->siteConfig->get('name');
      $params['seller_country'] = $this->country;
    }
    else {
      $props['payee'] = intertrading_wallet_id();
      $props['payer'] = $wid;
      $holder = Wallet::load($props['payer'])->getHolder();
      $params['buyer_name'] = $holder->label();
      $params['buyer_email'] = $holder->getEmail();
      $params['buyer_xname'] = $this->siteConfig->get('name');
      $params['buyer_country'] = $this->country;
    }
    $transaction = Transaction::create($props);

    $transaction->remote = (object)[
      'txid' => $params['txid'],
      'exchange_id' => $other_exchange_id,
      'user_id' => $other_exchange_name,
      'user_name' => $other_user_name
      ];

    $violations = $transaction->validate();
    if (count($violations)) {
      foreach ($violations as $violation) {
        //skip some constraints
        if ($violation->getPropertyPath() == 'payee' and $violation->getConstraint() instanceOf CanPayIn) {
          continue;
        }
        elseif($violation->getPropertyPath() == 'payer' and $violation->getConstraint() instanceOf CanPayOut) {
          continue;
        }
        $this->logger->error('@string', ['@string' => $violation->getMessage()]);
      }
      $params['response'] = SELF::ERR_SYNTAX;
    }
    else {
      $params['response'] = SELF::SUCCESS;
      $transaction->save();
    }
    $this->accountSwitcher->switchBack();
  }

  /**
   * @param string $login
   * @param string $pass
   * @return type
   */
  public function ping(string $login, string $pass) {
     $data = array(
      'nid' => $login,
      'password' => $pass,
      'system' => 'cforge'
    );
    try {
      $url = self::CLEARING_CENTRAL_URL.'/logon.php?'.http_build_query($data);
      $this->logger->debug('@url', ['@url' => $url]);
      $result = $this->httpClient->post($url);
    }
    catch (RequestException $e) {
      $this->messenger->addError($e->getMessage());
      return [];
    }
    $contents = $result->getBody()->getContents();
    // Need to distinguish between success and failed login.
    return $result->getStatusCode() == 200;

  }

  /**
   * {@inheritdoc}
   */
  public static function responseLookup(int $code) {
    $codes = [
      SELF::FAILURE => t('Unknown error in Clearing Central'),
      SELF::ERR_INVALID_USER => t('Buyer does not exist (unknown account number)'),
      SELF::ERR_EXCHANGE => t('Exchange not registered on Clearing Central'),
      SELF::ERR_DENIED => t('Transaction denied ("no funds", over limit, account locked, exchange over deficit limit etc.)'),
      SELF::ERR_SYNTAX => t('faulty data'),
      SELF::ERR_DUPLICATE => t('Repeat transaction and so rejected by CC (same TXID submitted)'),
      SELF::ERR_URL => t('URL error'),
      SELF::ERR_CONVERSION => t('Conversion rate not set'),
      SELF::ERR_SERVER => t('Server error (e.g. cannot access db)'),
      SELF::ERR_AUTH => t('Password is wrong in settings'),
      SELF::ERR_UNKNOWN_IP => t('IP of incoming server not in CC DB'),
      SELF::ERR_MISSING_TXID => t('No TXID provided (update/delete only)'),
      SELF::ERR_UNKNOWN_TXID => t('TXID does not exist (update/delete only)'),
      SELF::ERR_REMOTE => t('Unable to connect to remote server'),
      SELF::ERR_PEER_FAILED => t('Failed to connect to peer')
    ];
    return $codes[$code];
  }

  /**
   * Convert a transaction entity to a Clearing central remote transaction array
   *
   * @param TransactionInterface $transaction
   *
   * @return array
   *
   * @todo make this work for multiple exchanges
   */
  private function prepareToSend(TransactionInterface $transaction, $outgoing, $remote_exchange_id, $remote_user_id) {
    $wid = intertrading_wallet_id();
    $CcTransaction = [
      'txid' => $transaction->uuid->value,
      'description' => $transaction->description->value,
      'outgoing' => (int)$outgoing
    ];
    $payer = $transaction->payer->entity;
    $payee = $transaction->payee->entity;

    $intertrading_wallet = intertrading_wallet_id();

    if ($payee->id() == $intertrading_wallet) { // outgoing
      $CcTransaction += [
        'buyer_nid' => $this->exchangeId,
        'password' => $this->pass,
        'buyer_id' => $payer->id(),
        'buyer_name' => $payer->label(),
        'seller_nid' => $remote_exchange_id,
        'seller_id' => $remote_user_id,
        'amount' => $transaction->amount
      ];
    }
    elseif ($payer->id() == $intertrading_wallet) { // not outgoing
      $CcTransaction += [
        'seller_nid' => $this->exchangeId,
        'password' => $this->pass,
        'seller_id' => $payee->id(),
        'seller_name' => $payee->label(),
        'buyer_nid' => $remote_exchange_id,
        'buyer_id' => $remote_user_id,
        'amount' => $transaction->worth->format(CurrencyInterface::DISPLAY_PLAIN)
      ];
    }
    else {
      throw new \Exception("Either wallet $intertrading_wallet is not in transaction");
    }
    return $CcTransaction;
  }

  /**
   * Determine which wallet a fragment of text refers to. If its more than wallet, throw an error.
   * @param type $fragment
   * @return int
   *   The wallet ID, zero if no wallet matches
   *
   * @note mysql searches are case insensitive, unless the table collation is
   * latin1, in which case I don't know what to do
   */
  private function walletIdfromFragment($fragment) : int {
    $email = FALSE;
    $wid = 0;
    if (is_numeric($fragment)) {
      $wids = [$fragment];
    }
    // look for a matching email address.
    elseif (filter_var($fragment, FILTER_VALIDATE_EMAIL)) {
      $uids = \Drupal::entityQuery('user')
        ->condition('mail', $fragment)
        ->condition('status', TRUE)
        ->execute();
      if (count($uids) == 1) {
        $this->localWalletOwner = User::load(reset($uids));
        $wids = WalletStorage::walletsOf($this->localWalletOwner, FALSE);
      }
    }
    else {
      // Not an email, look for ONE similar wallet names.
      $wids = $this->walletEntityQuery->checkAccess(TRUE)
      ->condition('name', '%'.$fragment.'%', 'LIKE')
      ->execute();
      if (count($wids) > 1) {
        $this->logger->error(
          'Found more than one wallet matching fragment %fragment',
          ['%fragment' => $fragment]
        );
        return 0;
      }
      // last try: look for a matching user name, and get any wallet.
      if (empty($wids)) {
        $uids = \Drupal::entityQuery('user')
          ->condition('name', $fragment)
          ->condition('status', TRUE)
          ->execute();
        if (count($uids) == 1) {
          $this->localWalletOwner = User::load(reset($uids));
          $wids = WalletStorage::walletsOf($this->localWalletOwner);
        }
      }
    }

    return reset($wids);
  }

  /**
   * {@inheritdoc}
   */
  static function nidSearchUrl() {
    return Self::CLEARING_CENTRAL_URL . '/nidsearch.php';
  }

}

==> mutual_credit-5.0.6.tar.gz: mutual_credit/modules/clearingcentral/src/ClearingCentral.php <==
<?php

namespace Drupal\mcapi_cc;

use Drupal\mcapi\Entity\TransactionInterface;
use Drupal\mcapi\Entity\Wallet;
use Drupal\mcapi\Entity\Storage\WalletStorage;
use Drupal\mcapi\Plugin\Validation\Constraint\CanPayIn;
use Drupal\mcapi\Plugin\Validation\Constraint\CanPayOut;
use Drupal\mcapi\Entity\CurrencyInterface;
use Drupal\mcapi\Entity\Currency;
use Drupal\mcapi\Entity\Transaction;
use Drupal\user\Entity\User;
use Drupal\Core\Session\AccountSwitcher;
use Drupal\Core\Logger\LoggerChannel;
use Drupal\Core\Entity\EntityTypeManager;
use Drupal\Core\KeyValueStore\KeyValueFactory;
use Drupal\Core\Config\ConfigFactory;
use Drupal\Core\Messenger\MessengerInterface;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\RequestException;

/**
 * Manage communications with clearing central web service
 */
class ClearingCentral implements IntertradingInterface {

  protected $nameLookup;
  protected $httpClient;
  protected $logger;
  protected $country;
  protected $accountSwitcher;
  protected $messenger;
  protected $walletEntityQuery;

  private $exchangeId;
  private $password;
  private $currencyId;

  /**
   * @var Drupal\user\Entity\User
   */
  private $localWalletOwner;


  const CLEARING_CENTRAL_URL = 'http://clearingcentral.net';
  const CLEARING_CENTRAL_IP = '69.61.35.151';

  const SUCCESS = 1;
  const FAILURE = 0;
  const ERR_INVALID_USER = 2; //user (buyer) does not exist (unknown account number)
  const ERR_EXCHANGE = 3;
  const ERR_DENIED = 4; //"no funds", over limit, account locked, exchange over deficit limit etc.
  const ERR_SYNTAX= 5; // Faulty data
  const ERR_DUPLICATE= 6; //same TXID submitted
  const ERR_URL = 7;
  const ERR_CONVERSION = 8;
  const ERR_SERVER = 9;
  const ERR_AUTH = 10;
  const ERR_UNKNOWN_IP = 11;
  const ERR_MISSING_TXID = 12; //update/delete only
  const ERR_UNKNOWN_TXID = 13; //update/delete only
  const ERR_REMOTE = 14; //Unable to connect to remote server
  const ERR_PEER_FAILED = 17; //Peer failed verification


  /**
   * @param ConfigFactory $config_factory
   * @param ClientInterface $http_client
   * @param KeyValueFactory $key_value
   * @param EntitytypeManager $entity_type_manager
   * @param LoggerChannel $logger_channel
   * @param AccountSwitcher $account_switcher
   * @param MessengerInterface $messenger
   */
  function __construct(ConfigFactory $config_factory, ClientInterface $http_client, KeyValueFactory $key_value, EntityTypeManager $entity_type_manager, LoggerChannel $logger_channel, AccountSwitcher $account_switcher, MessengerInterface $messenger) {
    $this->siteConfig = $config_factory->get('system.site');
    $this->httpClient  = $http_client;
    $this->nameLookup  = $key_value->get('exchangeNames');
    $this->walletEntityQuery = $entity_type_manager->getStorage('mcapi_wallet')->getQuery();
    $this->logger = $logger_channel;
    $this->country = $config_factory->get('system.date')->get('country.default');
    $this->accountSwitcher = $account_switcher;
    $this->messenger = $messenger;

    $settings = $config_factory->get('mcapi_cc.settings');
    $this->exchangeId = $settings->get('login');
    $this->pass = $settings->get('pass');
    $this->currencyId = $settings->get('curr_id');
  }

  /**
   * {@inheritdoc}
   */
  public function send(TransactionInterface $transaction, $remote_exchange_id, $remote_user_id) : float {
    $remote_transaction = $this->prepareToSend(
      $transaction,
      $transaction->outgoing,
      $remote_exchange_id,
      $remote_user_id
    );
    $this->logger->debug('<pre>@array</pre>', ['@array' =>  print_r($remote_transaction, 1)]);
    try {
      $url = self::CLEARING_CENTRAL_URL.'/txinput.php?'.http_build_query($remote_transaction);
      $this->logger->debug('@url', ['@url' => $url]);
      $result = $this->httpClient->post($url);
    }
    catch (RequestException $e) {
      $this->messenger->addError($e->getMessage());
      return 0;
    }

    if ($result->getStatusCode() == 200) {
      $contents = $result->getBody()->getContents();
      parse_str($contents, $params);
      $response_code = (int)$params['response'];
      switch ($response_code) {
        case SELF::SUCCESS:
          // These are needed in hook_mcapi_transaction_insert.
          $transaction->remote->exchange_id = $remote_exchange_id;
          $transaction->remote->user_id = $remote_user_id;
          return $params['amount'];
        default:
          $this->messenger->addError($this->responseLookup($response_code));
          break;
      }
    }
    else {
      $message = t('Clearing central failed to respond.');
      $this->logger->error('@string', ['@string' => $message]);
      $this->messenger->addError($message);
      $form_state->setRebuild(TRUE);
    }
    return 0;
  }


  /**
   * {@inheritdoc}
   */
  public function receive(array &$params) : void {
    $props = [
      'uuid' => $params['txid'],
      'type' => 'remote',
      'description' => html_entity_decode($params['description'])
    ];

    if ($params['outgoing']) {
      //payment is going from the originating exchange to this exchange
      $local_exchange_id = $params['seller_nid'];
      $other_exchange_id = $params['buyer_nid'];
      $other_exchange_name = $params['buyer_xname'];
      $other_user_name = $params['buyer_name'];
      $wid = $this->walletIdFromfragment($params['seller_id']);
    }
    else {
      //payment is going from this exchange to the originating exchange
      $local_exchange_id = $params['buyer_nid'];
      $other_exchange_id = $params['seller_nid'];
      $other_exchange_name = $params['seller_xname'];
      $other_user_name = $params['seller_name'];
      $wid = $this->walletIdFromfragment($params['buyer_id']);
    }
    if (!$wid) {
      $this->logger->error(
        t(
          'Unable to find one wallet for fragment @frag',
          ['@frag' => $params['buyer_id']]
        )
      );
      $params['response'] = SELF::ERR_INVALID_USER;
      return;
    }
    else {
      // Replace the given seller_id with the actual wallet name.
      $params['seller_id'] = Wallet::load($wid)->name->value;
    }
    //store this for when we need to view the transaction
    if ($other_exchange_name) {
      $this->nameLookup->set($other_exchange_id, $other_exchange_name);
    }
    list($parts[1], $parts[3]) = explode('.', number_format($params['amount'], 2));

    $props['worth'] = [
      [
        'curr_id' => $this->currencyId,
        'value' => Currency::load($this->currencyId)->unformat($parts)
      ]
    ];
    $props['creator'] = 0;

    // wallet owner was created above in walletIdFromfragment()
    $this->accountSwitcher->switchTo($this->localWalletOwner);

    if ($params['outgoing']) {
      $props['payer'] = intertrading_wallet_id();
      $props['payee'] = $wid;
      $holder = Wallet::load($props['payee'])->getHolder();
      $params['seller_name'] = $holder->label();
      $params['seller_email'] = $holder->getEmail();
      $params['seller_xname'] = $this->siteConfig->get('name');
      $params['seller_country'] = $this->country;
    }
    else {
      $props['payee'] = intertrading_wallet_id();
      $props['payer'] = $wid;
      $holder = Wallet::load($props['payer'])->getHolder();
      $params['buyer_name'] = $holder->label();
      $params['buyer_email'] = $holder->getEmail();
      $params['buyer_xname'] = $this->siteConfig->get('name');
      $params['buyer_country'] = $this->country;
    }
    $transaction = Transaction::create($props);

    $transaction->remote = (object)[
      'txid' => $params['txid'],
      'exchange_id' => $other_exchange_id,
      'user_id' => $other_exchange_name,
      'user_name' => $other_user_name
      ];

    $violations = $transaction->validate();
    if (count($violations)) {
      foreach ($violations as $violation) {
        //skip some constraints
        if ($violation->getPropertyPath() == 'payee' and $violation->getConstraint() instanceOf CanPayIn) {
          continue;
        }
        elseif($violation->getPropertyPath() == 'payer' and $violation->getConstraint() instanceOf CanPayOut) {
          continue;
        }
        $this->logger->error(
          'Failed to validate transaction: @string',
          ['@string' => $violation->getMessage() . "<pre>\n".print_r($transaction->toArray(), 1).'<pre>']
        );
      }
      $params['response'] = SELF::ERR_SYNTAX;
    }
    else {
      $params['response'] = SELF::SUCCESS;
      $transaction->save();
    }
    $this->accountSwitcher->switchBack();
  }

  /**
   * @param string $login
   * @param string $pass
   * @return type
   */
  public function ping(string $login, string $pass) {
     $data = array(
      'nid' => $login,
      'password' => $pass,
      'system' => 'cforge'
    );
    try {
      $url = self::CLEARING_CENTRAL_URL.'/logon.php?'.http_build_query($data);
      $this->logger->debug('@url', ['@url' => $url]);
      $result = $this->httpClient->post($url);
    }
    catch (RequestException $e) {
      $this->messenger->addError($e->getMessage());
      return [];
    }
    $contents = $result->getBody()->getContents();
    // Need to distinguish between success and failed login.
    return $result->getStatusCode() == 200;

  }

  /**
   * {@inheritdoc}
   */
  public static function responseLookup(int $code) {
    $codes = [
      SELF::FAILURE => t('Unknown error in Clearing Central'),
      SELF::ERR_INVALID_USER => t('Buyer does not exist (unknown account number)'),
      SELF::ERR_EXCHANGE => t('Exchange not registered on Clearing Central'),
      SELF::ERR_DENIED => t('Transaction denied ("no funds", over limit, account locked, exchange over deficit limit etc.)'),
      SELF::ERR_SYNTAX => t('faulty data'),
      SELF::ERR_DUPLICATE => t('Repeat transaction and so rejected by CC (same TXID submitted)'),
      SELF::ERR_URL => t('URL error'),
      SELF::ERR_CONVERSION => t('Conversion rate not set'),
      SELF::ERR_SERVER => t('Server error (e.g. cannot access db)'),
      SELF::ERR_AUTH => t('Password is wrong in settings'),
      SELF::ERR_UNKNOWN_IP => t('IP of incoming server not in CC DB'),
      SELF::ERR_MISSING_TXID => t('No TXID provided (update/delete only)'),
      SELF::ERR_UNKNOWN_TXID => t('TXID does not exist (update/delete only)'),
      SELF::ERR_REMOTE => t('Unable to connect to remote server'),
      SELF::ERR_PEER_FAILED => t('Failed to connect to peer')
    ];
    return $codes[$code];
  }

  /**
   * Convert a transaction entity to a Clearing central remote transaction array
   *
   * @param TransactionInterface $transaction
   *
   * @return array
   *
   * @todo make this work for multiple exchanges
   */
  private function prepareToSend(TransactionInterface $transaction, $outgoing, $remote_exchange_id, $remote_user_id) {
    $wid = intertrading_wallet_id();
    $CcTransaction = [
      'txid' => $transaction->uuid->value,
      'description' => $transaction->description->value,
      'outgoing' => (int)$outgoing
    ];
    $payer = $transaction->payer->entity;
    $payee = $transaction->payee->entity;

    $intertrading_wallet = intertrading_wallet_id();

    if ($payee->id() == $intertrading_wallet) { // outgoing
      $CcTransaction += [
        'buyer_nid' => $this->exchangeId,
        'password' => $this->pass,
        'buyer_id' => $payer->id(),
        'buyer_name' => $payer->label(),
        'seller_nid' => $remote_exchange_id,
        'seller_id' => $remote_user_id,
        'amount' => $transaction->amount
      ];
    }
    elseif ($payer->id() == $intertrading_wallet) { // not outgoing
      $CcTransaction += [
        'seller_nid' => $this->exchangeId,
        'password' => $this->pass,
        'seller_id' => $payee->id(),
        'seller_name' => $payee->label(),
        'buyer_nid' => $remote_exchange_id,
        'buyer_id' => $remote_user_id,
        'amount' => $transaction->worth->format(CurrencyInterface::DISPLAY_PLAIN)
      ];
    }
    else {
      throw new \Exception("Either wallet $intertrading_wallet is not in transaction");
    }
    return $CcTransaction;
  }

  /**
   * Determine which wallet a fragment of text refers to. If its more than wallet, throw an error.
   * @param string $fragment
   * @return int
   *   The wallet ID, zero if no wallet matches
   *
   * @note mysql searches should case insensitive depending on the charset collation of the field.
   */
  private function walletIdfromFragment(string $fragment) : int {
    $email = FALSE;
    $wid = 0;
    if (is_numeric($fragment)) {
      $wids = [$fragment];
    }
    // look for a matching email address.
    elseif (filter_var($fragment, FILTER_VALIDATE_EMAIL)) {
      $uids = \Drupal::entityQuery('user')
        ->condition('mail', $fragment)
        ->condition('status', TRUE)
        ->execute();
      if (count($uids) == 1) {
        $this->localWalletOwner = User::load(reset($uids));
        $wids = WalletStorage::walletsOf($this->localWalletOwner, FALSE);
      }
    }
    else {
      // Not an email, look for ONE similar wallet names.
      $wids = $this->walletEntityQuery->checkAccess(TRUE)
      ->condition('name', '%'.$fragment.'%', 'LIKE')
      ->execute();
      if (count($wids) > 1) {
        $this->logger->error(
          'Found more than one wallet matching fragment %fragment',
          ['%fragment' => $fragment]
        );
        return 0;
      }
      // last try: look for a matching user name, and get any wallet.
      if (empty($wids)) {
        $uids = \Drupal::entityQuery('user')
          ->condition('name', $fragment)
          ->condition('status', TRUE)
          ->execute();
        if (count($uids) == 1) {
          $this->localWalletOwner = User::load(reset($uids));
          $wids = WalletStorage::walletsOf($this->localWalletOwner);
        }
      }
    }
    if (!$this->localWalletOwner and $wids) {
      $this->localWalletOwner = Wallet::load(reset($wids))->getOwner();
    }
    elseif (empty($wids)) {
      throw new \Exception('Unable to determine wallet owner');
    }
    return reset($wids);
  }

  /**
   * {@inheritdoc}
   */
  static function nidSearchUrl() {
    return Self::CLEARING_CENTRAL_URL . '/nidsearch.php';
  }

}
